----------------------------------
-- Variables
----------------------------------
TCC = {}
LAM = LibStub('LibAddonMenu-1.0')

-- Constants
TCC.NAME = 'TsChatColorizer'
TCC.SV = 'TCC_SV'
TCC.MAX_CUSTOM_COLOR_COUNT = 21

-- AddOn settings
TCC.settingsMenu = nil
TCC.settings = {
	customColorCount = 0,
	customColors = {}
}

-- AddOn settings color pickers
TCC.ColorPickers = {
	colorPickersCount = 0,
	colorPickers = {}
}

-- Color selection buttons
TCC.ColorSelectionButtons = {
	settings = {
		marginLeftRight = 6,
		marginTop = 4,
		
		columnSpacing = 5,
		rowSpacing = 5,
		
		width = 40,
		height = 22
	},
	
	presetButtonCount = 3,
	presetButtons = {
		removeAllColors = nil,
		setDefaultColor = nil,
		setColorRainbow = nil
	},
	
	customButtonCount = 0,
	customButtons = {}
}

-- Window properties
TCC.chatWindowWidth = 0

----------------------------------
-- Functions
----------------------------------
function TCC.OnInitialized()
	-- Register events
	EVENT_MANAGER:RegisterForEvent(TCC.NAME, EVENT_ADD_ON_LOADED, TCC.OnEventAddOnLoaded)

	TCC.RegisterHandlers()
end

TCC.handlerBackups = {
	ZO_ChatWindow = {},
	ZO_ChatWindowTextEntryLabel = {},
	ZO_ChatWindowTextEntryEditBox = {}
}

function TCC.RegisterHandlers()
	-- Backup the original handler and setup our handler and the default one
	-- Update the button positions when the chat window resizes
	if (ZO_ChatWindow:IsHandlerSet("OnResizeStart")) then
		TCC.handlerBackups.ZO_ChatWindow["OnResizeStart"] = ZO_ChatWindow:GetHandler("OnResizeStart")
		ZO_ChatWindow:SetHandler("OnResizeStart" , function(self) TCC.handlerBackups.ZO_ChatWindow["OnResizeStart"](self) TCC.OnChatWindowResizeStart() end)
	else
		ZO_ChatWindow:SetHandler("OnResizeStart" , function(self) TCC.OnChatWindowResizeStart() end)
	end
	
	if (ZO_ChatWindow:IsHandlerSet("OnResizeStop")) then
		TCC.handlerBackups.ZO_ChatWindow["OnResizeStop"] = ZO_ChatWindow:GetHandler("OnResizeStop")
		ZO_ChatWindow:SetHandler("OnResizeStop" , function(self) TCC.handlerBackups.ZO_ChatWindow["OnResizeStop"](self) TCC.OnChatWindowResizeStop() end)
	else
		ZO_ChatWindow:SetHandler("OnResizeStop" , function(self) TCC.OnChatWindowResizeStop() end)
	end
	
	-- Update the color button default color
	if (ZO_ChatWindowTextEntryLabel:IsHandlerSet("OnTextChanged")) then
		TCC.handlerBackups.ZO_ChatWindowTextEntryLabel["OnTextChanged"] = ZO_ChatWindowTextEntryLabel:GetHandler("OnTextChanged")
		ZO_ChatWindowTextEntryLabel:SetHandler("OnTextChanged" , function(self) TCC.handlerBackups.ZO_ChatWindowTextEntryLabel["OnTextChanged"](self) TCC.SetButtonColorDefaultColor() end)
	else
		ZO_ChatWindowTextEntryLabel:SetHandler("OnTextChanged" , function(self) TCC.SetButtonColorDefaultColor() end)
	end
	
	-- backspace handler to remove the last color
	if (ZO_ChatWindowTextEntryEditBox:IsHandlerSet("OnBackspace")) then
		TCC.handlerBackups.ZO_ChatWindowTextEntryEditBox["OnBackspace"] = ZO_ChatWindowTextEntryEditBox:GetHandler("OnBackspace")
		ZO_ChatWindowTextEntryEditBox:SetHandler("OnBackspace", function(self) TCC.handlerBackups.ZO_ChatWindowTextEntryEditBox["OnBackspace"](self) TCC.OnBackspaceRemoveColorFromText() end)
	else
		ZO_ChatWindowTextEntryEditBox:SetHandler("OnBackspace", function(self) TCC.OnBackspaceRemoveColorFromText() end)
	end
end

function TCC.OnEventAddOnLoaded(event, addonName)
	-- If the addon is not ours, do nothing
	if (addonName ~= TCC.NAME) then
		return
	end
	
	-- Intialize settings
	TCC.settings = ZO_SavedVars:NewAccountWide(TCC.SV, 1, nil, TCC.settings)
	
	-- Create the settings section with the help of LibAddonMenu and get the color settings
	TCC.InitializeSettings()
	
	-- Create the Color buttons and set their colors
	TCC.UpdateColorSelectionButtons()
	TCC.UpdateColorSelectionButtonColors()
end

function TCC.OnUpdate()
	TCC.RepositionColorButtonsWhileResizing()
	TCC.ApplyRainbowColorToText()
end

TCC.chatWindowResizingVars = {
	isChatWindowResizing = false,
	deltaTimer = 0
}
function TCC.RepositionColorButtonsWhileResizing()
	-- Reposition color buttons while resizing every 30 ms
	if ((GetGameTimeMilliseconds() - TCC.chatWindowResizingVars.deltaTimer) > 30) then
		if ((TCC.chatWindowResizingVars.isChatWindowResizing == true)) then
			TCC.chatWindowResizingVars.deltaTimer = GetGameTimeMilliseconds()
			
			TCC.UpdateColorSelectionButtonsPositions()
		end
	end
end

function TCC.OnChatWindowResizeStart()
	TCC.chatWindowResizingVars.isChatWindowResizing = true
end

function TCC.OnChatWindowResizeStop()
	TCC.chatWindowResizingVars.isChatWindowResizing = false
end
	
function TCC.InitializeSettings()
	-- Create the menu
	TCC.settingsMenu = LAM:CreateControlPanel( 'TCC_SettingsMenu', 'T\'s Chat Colorizer' )
	
	-- Header for the custom Color settings
	LAM:AddHeader(TCC.settingsMenu, "TCC_SettingsMenu_CustomColors", "Custom Colors")
	
	-- Add the slider that defines the custom Color button count
	local ColorsSlider = LAM:AddSlider(
		TCC.settingsMenu, 'TCC_SettingsMenu_CustomColors_Slide',
		'Colors', 'Sets the number of custom Colors to use.',
		0, TCC.MAX_CUSTOM_COLOR_COUNT, 1,
		function() return TCC.settings.customColorCount end,
		function(value)
			-- On a slider change, update the settings,
			-- add/remove buttons accordingly 
			-- and update the button positions
			TCC.settings.customColorCount = value
			TCC.UpdateSettingsColorPickers()
			TCC.UpdateColorSelectionButtons()
			TCC.UpdateColorSelectionButtonsPositions()
		end
	)
	
	-- Add the color pickers
	TCC.UpdateSettingsColorPickers()
end

function TCC.CustomColorButtonOnMouseEnter(button)
	button.highlight:SetAlpha(0.35)
end

function TCC.CustomColorButtonOnMouseExit(button)
	button.highlight:SetAlpha(0)
end

function TCC.CustomColorButtonOnMouseDown(button)
	if (TCC.TextCustomization.selected ~= nil and TCC.TextCustomization.selected == button) then
		button:GetChild(1):SetScale(1.0)
		button:GetChild(2):SetScale(1.0)
		button:GetChild(3):SetScale(1.0)
	else
		button:GetChild(1):SetScale(0.9)
		button:GetChild(2):SetScale(0.9)
		button:GetChild(3):SetScale(0.9)
	end
end

function TCC.CustomColorButtonOnMouseUp(button)
	TCC.TextCustomization.rainbow = false
	
	if (TCC.TextCustomization.selected ~= nil) then
		TCC.TextCustomization.selected:GetChild(1):SetScale(1.0)
		TCC.TextCustomization.selected:GetChild(2):SetScale(1.0)
		TCC.TextCustomization.selected:GetChild(3):SetScale(1.0)
	end
	
	if (string.match(button:GetName(), "Remove")) then
		TCC.TextCustomization.selected = nil
		button:GetChild(1):SetScale(1.0)
		button:GetChild(2):SetScale(1.0)
		button:GetChild(3):SetScale(1.0)
	else
		button:GetChild(1):SetScale(1.2)
		button:GetChild(2):SetScale(1.2)
		button:GetChild(3):SetScale(1.2)
		TCC.TextCustomization.selected = button
	end

	if (string.match(button:GetName(), "Remove")) then
		TCC.RemoveAllColorFromText()
	elseif (string.match(button:GetName(), "Default")) then
		local r, g, b = button.bg:GetColor()
		TCC.TextCustomization.currentColor = TCC.GetTextColorCodeFromRGB(r, g, b)
		TCC.ApplyColorToText()
	elseif (string.match(button:GetName(), "Rainbow")) then
		-- remove the last color if there is one
		local currentText = ZO_ChatWindowTextEntryEditBox:GetText()
		if (string.utf8sub(currentText, string.utf8len(currentText)-7, string.utf8len(currentText)-6) == "|c") then
			currentText = string.utf8sub(currentText, 1, string.utf8len(currentText)-8)
		end
		
		TCC.TextCustomization.rainbow = true
	else
		local r, g, b = button.bg:GetColor()
		TCC.TextCustomization.currentColor = TCC.GetTextColorCodeFromRGB(r, g, b)
		TCC.ApplyColorToText()
	end
end

-- Text customization
TCC.TextCustomization = {
	oldTextForRainbow = "",
	rainbow = false,
	currentColor = "",
	selected = nil,
	lastKeyIsBackspace = false,
	
	rainbowColor = {
		currentIndex = 1,
		colorCount = 12,
		color = {
			[1] = "|cff0000",
			[2] = "|cff008c",
			[3] = "|cff00ff",
			[4] = "|c7b00ff",
			[5] = "|c0000ff",
			[6] = "|c0060ff",
			[7] = "|c00ffff",
			[8] = "|c00ff88",
			[9] = "|c00ff00",
			[10] = "|c84ff00",
			[11] = "|cffff00",
			[12] = "|cff8100"
		}
	}
}

function TCC.ApplyColorToText(overrideColor)
	local color = TCC.TextCustomization.currentColor
	local currentText = ZO_ChatWindowTextEntryEditBox:GetText()

	-- remove the last color if there is one
	if (string.utf8sub(currentText, string.utf8len(currentText)-7, string.utf8len(currentText)-6) == "|c") then
		currentText = string.utf8sub(currentText, 1, string.utf8len(currentText)-8)
	end
	
	currentText = currentText .. color
	ZO_ChatWindowTextEntryEditBox:SetText(currentText)
end

function TCC.ApplyRainbowColorToText()
	if ((TCC.TextCustomization.rainbow == true) and ZO_ChatWindowTextEntryEditBox:HasFocus()) then
		local currentText = ZO_ChatWindowTextEntryEditBox:GetText()
		if (TCC.TextCustomization.oldTextForRainbow ~= currentText) then
			if (TCC.TextCustomization.lastKeyIsBackspace) then
				TCC.TextCustomization.lastKeyIsBackspace = false
				TCC.TextCustomization.oldTextForRainbow = currentText
			else
				-- place the color before the key that was just entered
				local lastChar = string.utf8sub(currentText, string.utf8len(currentText), string.utf8len(currentText))
				local restText = string.utf8sub(currentText, 1, string.utf8len(currentText)-1)
				
				local newText = restText .. TCC.TextCustomization.rainbowColor.color[TCC.TextCustomization.rainbowColor.currentIndex] .. lastChar
				ZO_ChatWindowTextEntryEditBox:SetText(newText)
				
				if (TCC.TextCustomization.rainbowColor.currentIndex == TCC.TextCustomization.rainbowColor.colorCount) then
					TCC.TextCustomization.rainbowColor.currentIndex = 1
				else
					TCC.TextCustomization.rainbowColor.currentIndex = TCC.TextCustomization.rainbowColor.currentIndex + 1
				end
				
				TCC.TextCustomization.oldTextForRainbow = newText
			end
		end
	end
end

function TCC.OnBackspaceRemoveColorFromText()
	local currentText = ZO_ChatWindowTextEntryEditBox:GetText()
	
	if (string.utf8sub(currentText, string.utf8len(currentText)-6, string.utf8len(currentText)-5) == "|c") then
		currentText = string.utf8sub(currentText, 1, string.utf8len(currentText)-8)
	end
	
	ZO_ChatWindowTextEntryEditBox:SetText(currentText)
	
	TCC.TextCustomization.lastKeyIsBackspace = true
end

function TCC.RemoveAllColorFromText()
	local currentText = ZO_ChatWindowTextEntryEditBox:GetText()
	
	-- remove all color codes
	local found = string.find(currentText, "|c")
	while found ~= nil do
		local first = string.utf8sub(currentText, 1, found-1)
		local last = string.utf8sub(currentText, found+8, string.utf8len(currentText))
	
		currentText = first .. last
		found = string.find(currentText, "|c")
	end

	ZO_ChatWindowTextEntryEditBox:SetText(currentText)
end

function TCC.UpdateSettingsColorPickers()
	if (TCC.settings.customColorCount > TCC.ColorPickers.colorPickersCount) then
		-- Add Color pickers
		for i = TCC.ColorPickers.colorPickersCount + 1, TCC.settings.customColorCount, 1 do
			-- Create it
			if (TCC.ColorPickers.colorPickers[i] == nil) then
				TCC.ColorPickers.colorPickers[i] = LAM:AddColorPicker(
					TCC.settingsMenu, 'TCC_SettingsMenu_CustomColors_Color' .. i,
					'     Custom Color ' .. i, 'Select a custom Color.',
					function() if (TCC.settings.customColors[i] == nil) then TCC.settings.customColors[i] = {r = 1.0, g = 1.0, b = 1.0} end return TCC.settings.customColors[i].r, TCC.settings.customColors[i].g, TCC.settings.customColors[i].b end,
					function(r, g, b) TCC.settings.customColors[i].r = r TCC.settings.customColors[i].g = g TCC.settings.customColors[i].b = b TCC.UpdateColorSelectionButtonColors() TCC.UpdateColorSelectionButtons() end
				)
			else
				-- Show it
				TCC.ColorPickers.colorPickers[i]:SetAlpha(1)
			end
		end
	elseif (TCC.settings.customColorCount < TCC.ColorPickers.colorPickersCount) then
		-- Remove Color pickers
		for i = TCC.ColorPickers.colorPickersCount, TCC.settings.customColorCount + 1, -1 do
			-- Hide it
			TCC.ColorPickers.colorPickers[i]:SetAlpha(0)
			
			-- Remove config entry
			TCC.settings.customColors[i] = nil
		end
	end
	
	TCC.ColorPickers.colorPickersCount = TCC.settings.customColorCount
end

function TCC.UpdateColorSelectionButtons()
	for i = 1, TCC.ColorSelectionButtons.presetButtonCount + TCC.settings.customColorCount, 1 do
		if (TCC.ColorSelectionButtons.presetButtons.removeAllColors == nil and i == 1) then
			-- Remove all color button
			TCC.ColorSelectionButtons.presetButtons.removeAllColors = TCC.CreateColorSelectionButton("Remove")
			TCC.ColorSelectionButtons.presetButtons.removeAllColors.bg:SetTexture("\\TsChatColorizer\\Textures\\remove.dds")
		elseif (TCC.ColorSelectionButtons.presetButtons.setDefaultColor == nil and i == 2) then
			-- Set to default color button
			TCC.ColorSelectionButtons.presetButtons.setDefaultColor = TCC.CreateColorSelectionButton("Default")
		elseif (TCC.ColorSelectionButtons.presetButtons.setColorRainbow == nil and i == 3) then
			-- Type in rainbow colors button
			TCC.ColorSelectionButtons.presetButtons.setColorRainbow = TCC.CreateColorSelectionButton("Rainbow")
			TCC.ColorSelectionButtons.presetButtons.setColorRainbow.bg:SetTexture("\\TsChatColorizer\\Textures\\rainbow.dds")
		end
	end
	
	if (TCC.settings.customColorCount > TCC.ColorSelectionButtons.customButtonCount) then
		for i = TCC.ColorSelectionButtons.customButtonCount + 1, TCC.settings.customColorCount, 1 do
			if (TCC.ColorSelectionButtons.customButtons[i] == nil) then
				-- Custom color selection buttons
				TCC.ColorSelectionButtons.customButtons[i] = TCC.CreateColorSelectionButton("Custom" .. i)
			else
				TCC.ColorSelectionButtons.customButtons[i]:SetAlpha(1)
			end
		end
	elseif (TCC.settings.customColorCount < TCC.ColorSelectionButtons.customButtonCount) then
		for i = TCC.ColorSelectionButtons.customButtonCount, TCC.settings.customColorCount + 1, -1 do
			TCC.ColorSelectionButtons.customButtons[i]:SetAlpha(0)
		end
	end
	
	TCC.ColorSelectionButtons.customButtonCount = TCC.settings.customColorCount
end

function TCC.CreateColorSelectionButton(postfix)
	local control = WINDOW_MANAGER:CreateControl("TCC_ChatWindow_CustomColor" .. postfix, ZO_ChatWindow, CT_CONTROL)
	control:SetDimensions(36, 18)
	control:SetAnchor(TOPLEFT, ZO_ChatWindow, BOTTOMLEFT, 0, 0)
	control:SetMouseEnabled(true)
	
	control.bg = WINDOW_MANAGER:CreateControl("TCC_ChatWindow_CustomColorBG" .. postfix, control, CT_TEXTURE)
	control.bg:SetColor(1, 1, 1, 1)
	control.bg:SetDimensions(36, 18)
	control.bg:SetAnchor(CENTER, control, CENTER, 0, 0)
	control.bg:SetDrawLayer(0)
	
	control.frame = WINDOW_MANAGER:CreateControl("TCC_ChatWindow_CustomColorFrame" .. postfix, control, CT_TEXTURE)
	control.frame:SetTexture("EsoUI\\Art\\ChatWindow\\chatOptions_bgColSwatch_frame.dds")
	control.frame:SetTextureCoords(0, .625, 0, .8125)
	control.frame:SetDimensions(40, 22)
	control.frame:SetAnchor(CENTER, control, CENTER, 0, 0)
	control.frame:SetDrawLayer(1)
	
	control.highlight = WINDOW_MANAGER:CreateControl("TCC_ChatWindow_CustomColorHighlight" .. postfix, control, CT_TEXTURE)
	control.highlight:SetColor(1, 1, 1, 1)
	control.highlight:SetDimensions(40, 22)
	control.highlight:SetAnchor(CENTER, control, CENTER, 0, 0)
	control.highlight:SetDrawLayer(2)
	control.highlight:SetAlpha(0)

	control:SetHandler("OnMouseEnter" , function(self) TCC.CustomColorButtonOnMouseEnter(self) end)
	control:SetHandler("OnMouseExit" , function(self) TCC.CustomColorButtonOnMouseExit(self) end)
	control:SetHandler("OnMouseUp" , function(self) TCC.CustomColorButtonOnMouseUp(self) end)
	control:SetHandler("OnMouseDown" , function(self) TCC.CustomColorButtonOnMouseDown(self) end)
	
	return control
end

function TCC.UpdateColorSelectionButtonsPositions()
	TCC.chatWindowWidth = ZO_ChatWindow:GetWidth()
	
	local currentRow = 0
	local currentColumn = 0
	local trueWindowWidth = TCC.chatWindowWidth - TCC.ColorSelectionButtons.settings.marginLeftRight * 2
	local maxColumnCount = math.ceil(trueWindowWidth /(TCC.ColorSelectionButtons.settings.width + TCC.ColorSelectionButtons.settings.columnSpacing)) - 1

	for i = 1, TCC.ColorSelectionButtons.presetButtonCount + TCC.settings.customColorCount, 1 do
		-- Caluclate the position
		local firstRowButtonCount = 0
		if (TCC.ColorSelectionButtons.presetButtonCount + TCC.settings.customColorCount < maxColumnCount) then
			firstRowButtonCount = TCC.ColorSelectionButtons.presetButtonCount + TCC.settings.customColorCount
		else
			firstRowButtonCount = maxColumnCount
		end
		
		local totalButtonsWidth = firstRowButtonCount * (TCC.ColorSelectionButtons.settings.width + TCC.ColorSelectionButtons.settings.columnSpacing) - TCC.ColorSelectionButtons.settings.columnSpacing
		local leftoverSpace = (trueWindowWidth - totalButtonsWidth) / 2
		local offsetX = currentColumn * (TCC.ColorSelectionButtons.settings.width + TCC.ColorSelectionButtons.settings.columnSpacing)
		local offsetY = currentRow * (TCC.ColorSelectionButtons.settings.height + TCC.ColorSelectionButtons.settings.rowSpacing)

		-- Apply the position
		if (i == 1) then
			TCC.ColorSelectionButtons.presetButtons.removeAllColors:SetAnchor(TOPLEFT, ZO_ChatWindow, BOTTOMLEFT, TCC.ColorSelectionButtons.settings.marginLeftRight + offsetX + leftoverSpace, TCC.ColorSelectionButtons.settings.marginTop + offsetY)
		elseif (i == 2) then
			TCC.ColorSelectionButtons.presetButtons.setDefaultColor:SetAnchor(TOPLEFT, ZO_ChatWindow, BOTTOMLEFT, TCC.ColorSelectionButtons.settings.marginLeftRight + offsetX + leftoverSpace, TCC.ColorSelectionButtons.settings.marginTop + offsetY)
		elseif (i == 3) then
			TCC.ColorSelectionButtons.presetButtons.setColorRainbow:SetAnchor(TOPLEFT, ZO_ChatWindow, BOTTOMLEFT, TCC.ColorSelectionButtons.settings.marginLeftRight + offsetX + leftoverSpace, TCC.ColorSelectionButtons.settings.marginTop + offsetY)
		elseif (i > TCC.ColorSelectionButtons.presetButtonCount) then
			TCC.ColorSelectionButtons.customButtons[i - TCC.ColorSelectionButtons.presetButtonCount]:SetAnchor(TOPLEFT, ZO_ChatWindow, BOTTOMLEFT, TCC.ColorSelectionButtons.settings.marginLeftRight + offsetX + leftoverSpace, TCC.ColorSelectionButtons.settings.marginTop + offsetY)
		end
		
		currentColumn = currentColumn + 1
		if (currentColumn == maxColumnCount) then
			currentColumn = 0
			currentRow = currentRow + 1
		end
	end
end

function TCC.UpdateColorSelectionButtonColors()
	for i = 1, TCC.ColorSelectionButtons.presetButtonCount + TCC.settings.customColorCount, 1 do
		if (i == 2) then
			TCC.SetButtonColorDefaultColor()
		elseif (i > TCC.ColorSelectionButtons.presetButtonCount) then
			local index = i - TCC.ColorSelectionButtons.presetButtonCount
			TCC.ColorSelectionButtons.customButtons[index].bg:SetColor(TCC.settings.customColors[index].r, TCC.settings.customColors[index].g, TCC.settings.customColors[index].b, 1)
		end
	end
end

TCC.buttonsPositionInitialized = false
function TCC.SetButtonColorDefaultColor()
	local r, g, b = ZO_ChatWindowTextEntryLabel:GetColor()
	TCC.ColorSelectionButtons.presetButtons.setDefaultColor.bg:SetColor(r, g, b, 1)
	
	-- Hacky way to get this to initialize properly,
	-- else the buttons will position themselfs before
	-- the window width is defined
	if (TCC.buttonsPositionInitialized == false and string.utf8len(ZO_ChatWindowTextEntryLabel:GetText()) > 0 ) then
		TCC.UpdateColorSelectionButtonsPositions()
		TCC.buttonsPositionInitialized = true
	end
end

function TCC.GetTextColorCodeFromRGB(r, g, b)
	return '|c' .. DEC_HEX(r*255) .. DEC_HEX(g*255) .. DEC_HEX(b*255)
end